! Renee modified this module in October 2010 to take in vector of k values "knots" and the corresponding
!  values of the bands (corresponding to the astro-ph/0302306 paper) over which we will spline

! This module provides the initial power spectra, parameterized as an expansion in ln k
!
! ln P = ln A_s + (n_s -1)*ln(k/k_0) + n_{run}/2 * ln(k/k_0)^2 
!
! so if n_run = 0
!
! P = A_s (k/k_0_scalar)^(n_s-1)
!
!for the scalar spectrum, when an(in) is the in'th spectral index. k_0_scalar
!is a pivot scale, fixed here to 0.05/Mpc (change it below as desired).
!
!This module uses the same inputs an(in), ant(in) and rat(in) as CMBFAST, however here
!rat(in) is used to set the ratio of the initial power spectra, so here
!
!** rat(in) is not the Cl quadrupole ratio ***
!
!in general models the quadrupole ratio depends in a complicated way on the ratio of the initial
!power spectra

!The absolute normalization of the Cls is unimportant here, but the relative ratio
!of the tensor and scalar Cls generated with this module will be correct for general models


!The OutputNormalization parameter controls the final output
!Absolute Cls can be obtained by setting OuputNormalization=outNone, otherwise the overall normalization
!of the power spectra doesn't matter

!This version December 2003 - changed default tensor pivot to 0.05 (consistent with CMBFAST 4.5)

     module InitialPower   
     use Precision
     implicit none   

     private
    
      character(LEN=*), parameter :: Power_Name = 'power_tilt'

      integer, parameter :: nnmax= 5
      integer, parameter :: knotmax = 50
       !Maximum possible number of different power spectra to use

      Type InitialPowerParams

       integer nn  !Must have this variable
       integer kn ! the number of knots
      !The actual number of power spectra to use
  
      !In this implementation we will use a vector for the P(k)
       real(dl) bvals(nnmax,knotmax) ! the bandpower values
       real(dl) kvals(nnmax,knotmax) ! the values of k at which the knots are placed
       real(dl) bvalsecondderivs(nnmax,knotmax)
       real(dl) rat(nnmax) !ratio of scalar to tensor initial power spectrum amplitudes
       real(dl) ScalarPowerAmp(nnmax) ! the overall amplitude 
       real(dl)  :: k_0_tensor
       real(dl) ant(nnmax)
       
      end Type InitialPowerParams

      real(dl) curv  !Curvature contant, set in InitializePowers     
      
      Type(InitialPowerParams) :: P
  
!Make things visible as neccessary...
 
      public InitialPowerParams, InitialPower_ReadParams, InitializePowers, ScalarPower, TensorPower
      public nnmax,Power_Descript, Power_Name, SetDefPowerParams
!      public 
    contains
       
  
       subroutine SetDefPowerParams(AP)
         real(dl) :: kval, bval
         integer i,j
         
         Type (InitialPowerParams) :: AP
        
         AP%nn     = 1 !number of initial power spectra
         AP%kn     = 20 !number of knots

         
         open(unit = 77, file='bvalDefault.dat', action = 'read', status = 'unknown')
         do i=1,AP%nn
            do j=1,AP%kn
               read(77,*) kval, bval
               
               AP%kvals(i,j) = kval
               AP%bvals(i,j) = bval
               !write(*,*), AP%kvals(i,j), AP%bvals(i,j)
         enddo
         enddo
         close(77)
         AP%ant    = 0 !tensor spectra index
         AP%rat    = 1
         AP%k_0_tensor = 0.05
         AP%ScalarPowerAmp = 2.36e-9
         

         do i=1,AP%nn
            call spline(AP%kvals(i,:),AP%bvals(i,:), AP%kn,0,0,AP%bvalsecondderivs(i,:))
         enddo
         
         
       end subroutine SetDefPowerParams

       subroutine InitializePowers(AParamSet,acurv)
         Type (InitialPowerParams) :: AParamSet
         !Called before computing final Cls in cmbmain.f90
         !Could read spectra from disk here, do other processing, etc.

        real(dl) acurv

        if (AParamSet%nn > nnmax) then
           write (*,*) 'To use ',AParamSet%nn,'power spectra you need to increase'
           write (*,*) 'nnmax in power_tilt.f90, currently ',nnmax
        end if
        P = AParamSet

        curv=acurv         

!Write implementation specific code here...

       end subroutine InitializePowers
       

       function ScalarPower(k, in)
         
         ! "in" is the index of the spectrum - in case there are more - this is related to AP%nn
         ! we reference all qualities by that now.. we will need to read off:
         ! kknots, bvalknots, nknots
         ! kvec is the vector of length lkvec which will be splined to get the Pk vector                       
         ! kknots and bvalknots are the vectors of k points and the corresponding values of b, which make the \
         ! "top hat" P(k)                                                                                            
         ! they are of length nknots                                                                           
         
         integer in
         integer :: j
         real(dl) :: ScalarPower, k,h
         integer :: kbin
         real(dl) :: a, b
         integer indlow,indhigh


         if (in > 1) then

            write(*,*)' ARGH THERE IS A PROBLEM RENEE'
            stop
         endif

         
         kbin = P%kn
         scalarpower=-1.0
         if (k<P%kvals(in,1)) then
                  !write(*,*) 'below lower bin'
            ScalarPower=1
         elseif (k .ge. P%kvals(in,P%kn)) then
            !write(*,*) 'above upper bin'
            ScalarPower = P%bvals(in,P%kn)
         else
            do j = 1,kbin-1 ! looking over the specifiec number of k bins
               if (k>P%kvals(in,j) .and. k<P%kvals(in,j+1)) then
                  !write(*,*) 'Used bin ',j
                  
                  h=P%kvals(in,j+1)-P%kvals(in,j)
                  if (h ==0) then
                     write(*,*) 'There is a problem in the spline over initial power - check the size of your bins'
                     stop
                  endif

                  a=(P%kvals(in,j+1)-k)/h
                  b=(k-P%kvals(in,j))/h
                  !write(*,*) a,b
                  ScalarPower =  (a*P%bvals(in,j) + b*P%bvals(in,j+1) + ((a**3-a)*P%bvalsecondderivs(in,j) + &
                       (b**3-b)*P%bvalsecondderivs(in,j+1))*(h**2)/6._dl)
                  exit
               endif
            enddo
         endif
         ScalarPower = P%ScalarPowerAmp(in)*ScalarPower
         !write(*,*) "XYZ", k, ScalarPower
         if (ScalarPower<0.0) then 
         
         write(*,*) 'Error setting k value - bins wrong in parameter file?  Should be ascending order.'
         stop
         endif
         
       end function ScalarPower
      
      function TensorPower(k,in)
      
       !TensorPower= const for scale invariant spectrum
       !The normalization is defined so that
       ! < h_{ij}(x) h^{ij}(x) > = \sum_nu nu /(nu^2-1) (nu^2-4)/nu^2 TensorPower(k)
       !for a closed model
       ! < h_{ij}(x) h^{ij}(x) > = int d nu /(nu^2+1) (nu^2+4)/nu^2 TensorPower(k)
       !for an open model
       !"in" gives the index of the power spectrum to return 
       !Here nu^2 = (k^2 + 3*curv)/|curv| 


        real(dl) TensorPower,k   
        real(dl), parameter :: PiByTwo=3.14159265d0/2._dl
   
        integer in

        TensorPower=P%rat(in)*P%ScalarPowerAmp(in)*exp(P%ant(in)*log(k/P%k_0_tensor))
        if (curv < 0) TensorPower=TensorPower*tanh(PiByTwo*sqrt(-k**2/curv-3)) 

       
      end function TensorPower

      !Get parameters describing parameterisation (for FITS file)
     function Power_Descript(in, Scal, Tens, Keys, Vals)
         character(LEN=8), intent(out) :: Keys(*)
         real(dl), intent(out) :: Vals(*)
         integer, intent(IN) :: in
         logical, intent(IN) :: Scal, Tens
         integer num, Power_Descript
         num=0
         if (Scal) then
            num=num+1
            Keys(num) = 'n_s'
            Vals(num) = 0!P%an(in)
            num=num+1
            Keys(num) = 'n_run'
            Vals(num) = 0!P%n_run(in)
            num=num+1
            Keys(num) = 's_pivot'
            Vals(num) = 0!P%k_0_scalar
         end if
         if (Tens) then
            num=num+1
            Keys(num) = 'n_t'
            Vals(num) = P%ant(in)
            num=num+1
            Keys(num) = 't_pivot'
            Vals(num) = P%k_0_tensor
            if (Scal) then
               num=num+1
               Keys(num) = 'p_ratio'
               Vals(num) = P%rat(in)
            end if
         end if
         Power_Descript = num

       end  function Power_Descript
        
       subroutine InitialPower_ReadParams(InitPower, Ini, WantTensors)
          use IniFile
          Type(InitialPowerParams) :: InitPower
          Type(TIniFile) :: Ini
          logical, intent(in) :: WantTensors
          integer i,j
          
           !InitPower%k_0_scalar = Ini_Read_Double_File(Ini,'pivot_scalar',InitPower%k_0_scalar)
           InitPower%k_0_tensor = Ini_Read_Double_File(Ini,'pivot_tensor',InitPower%k_0_tensor) 
           InitPower%nn = Ini_Read_Int('initial_power_num')
           InitPower%kn = Ini_Read_Int('initial_knot_num')
           
           if (InitPower%nn>nnmax) stop 'Too many initial power spectra - increase nnmax in InitialPower'
           InitPower%rat(:) = 1
           

           do j=1, InitPower%kn

              
                 InitPower%bvals(1,j)=Ini_Read_Double_Array_File(Ini, 'bandpower',j)
                 InitPower%kvals(1,j) = Ini_Read_Double_Array_File(Ini, 'knotpos',j)
                 !write(*,*) 'Inside InitialPower_ReadParams'
                 !write(*,*) InitPower%kvals(1,j), InitPower%bvals(1,j)
              enddo
              

              call spline(InitPower%kvals(1,:),InitPower%bvals(1,:), InitPower%kn,0,0,InitPower%bvalsecondderivs(1,:))

              do i = 1,InitPower%nn
                 if (WantTensors) then
                    InitPower%ant(i) = Ini_Read_Double_Array_File(Ini,'tensor_spectral_index',i)
                    InitPower%rat(i) = Ini_Read_Double_Array_File(Ini,'initial_ratio',i)
                 end if

              InitPower%ScalarPowerAmp(i) = Ini_Read_Double_Array_File(Ini,'scalar_amp',i,1._dl) 
              write(*,*) InitPower%ScalarPowerAmp(i)
              !Always need this as may want to set tensor amplitude even if scalars not computed
           end do
          
       end  subroutine InitialPower_ReadParams 


     end module InitialPower
