!Proposal density 
!Using the covariance matrix to give the proposal directions typically
!significantly increases the acceptance rate and gives faster movement 
!around parameter space.
!We generate a random basis in the eigenvectors, then cycle through them
!proposing changes to each, then generate a new random basis
!The distance proposal in the random direction is given by a two-D Gaussian 
!radial function mixed with an exponential, which is quite robust to wrong width estimates
!See http://cosmologist.info/notes/cosmomc.ps.gz

module propose
  use settings
  use Random
  use ParamDef
  implicit none

 logical :: propose_rand_directions = .false.
 logical :: fast_slicing = .false. !slice fast parameters when using Metropolis 
 logical :: slice_stepout = .false.
 logical :: slice_randomsize = .false.

 real, allocatable, dimension(:,:), save :: Rot_slow, Rot_fast

contains

 subroutine RotMatrix(M,n)
  integer, intent(in) :: n
  real M(n,n)
  integer i
  
   if (propose_rand_directions .and. n > 1) then
      
      call RandRotation(M, n)      
   else
      M = 0
      do i = 1, n
           M(i,i) = sign(1., real(ranmar())-0.5)
           ! taking a step either forward or backward along the direction - sign just changes the sign
      end do
   end if
 
 end  subroutine RotMatrix


 function Propose_r(in_n) result(r_fac)
  !distance proposal function (scale is 1)
   integer, intent(in) :: in_n
   integer i,n
   real r_fac

      if (ranmar() < 0.33d0) then
       r_fac = randexp1()
      else
       n = min(in_n,2)
       r_fac = 0
       do i = 1, n
        r_fac = r_fac + Gaussian1()**2
       end do
       r_fac = sqrt( r_fac / n )
      end if

 end function Propose_r

 function Propose_rRH(in_nrh) result(r_facrh)
  !distance proposal function (scale is 1)
   integer, intent(in) :: in_nrh
   integer i,n
   real r_facrh

       r_facrh = sqrt( 1.d0 / in_nrh )


 end function Propose_rRH


 subroutine UpdateParamsDirection(tmp, fast, dist,i)
  !Change parameters in tmp by dist in the direction of the ith random e-vector
   Type(ParamSet) tmp
   real vec(num_params_used)
   real distcheck
   integer, intent(in) :: i
   real  dummy(num_params_used), propose_amp
   real(8)  A(num_slow, num_slow)
   logical, intent(in) :: fast
   real, intent(in) :: dist

   integer :: j, jj, info


   propose_amp = 2.4d0
   ! NOTE that this assumes that the covariance matrix has already been square rooted previously!

   if (num==1)  then
      write(*,*) '********* first step **********' 
      tmp%P(:) = Scales%center(:)
   else
      

      do j = 1, num_slow
         vec(j) = Gaussian1()
      enddo


      tmp%P(params_used) =  tmp%P(params_used) + propose_amp/sqrt(dble(num_slow))*matmul (propose_matrix(1:num_slow,1:num_slow), vec(1:num_slow))

endif
      
 end subroutine UpdateParamsDirection


subroutine GetProposal(In, Out)

  type (ParamSet) In, Out
  integer, save :: fast_ix = 0

  fast_ix = fast_ix + 1
  if (num_slow ==0 .or. num_fast /= 0 .and. &
       mod(fast_ix, 2*(1 + (num_fast*oversample_fast)/num_slow)) /= 0) then
   call GetProposalProjFast(In, Out)
  else
   call GetProposalProjSlow(In, Out)
  end if

end subroutine GetProposal

subroutine GetProposalProjSlow(In, Out)
  use settings
  use Random
  use ParamDef
  implicit none
  type (ParamSet) In, Out
  real  wid
  integer, save :: loopix = 0
 
  slow_proposals = slow_proposals + 1
!Change only slow params, or eigenvectors of covmat which most change the slow params

   Out= In
   wid = propose_scale
   if (mod(loopix,num_slow)==0) then
        if (.not. allocated(Rot_slow)) allocate(Rot_slow(num_slow,num_slow))
        call RotMatrix(Rot_slow, num_slow)      
        loopix = 0
   end if
   loopix = loopix + 1

   call UpdateParamsDirection(Out,.false.,Propose_r(num_slow) * wid, loopix)

end subroutine GetProposalProjSlow


subroutine GetProposalProjFast(In, Out, ascale)
  type (ParamSet) In, Out
  real wid
  real, intent(in), optional :: ascale
  integer, save :: loopix = 0

   Out= In

   wid = propose_scale
   if (present(ascale)) wid = ascale

   if (mod(loopix,num_fast)==0) then
        if (.not. allocated(Rot_fast)) allocate(Rot_fast(num_fast,num_fast))
        call RotMatrix(Rot_fast, num_fast)      
        loopix = 0
   end if
   loopix = loopix + 1

   call UpdateParamsDirection(Out,.true.,Propose_r(num_fast) * wid, loopix)
     
end subroutine GetProposalProjFast

end module propose
